/* (C) 2004 by Dmytry Lavrov.
contact: dmytry lavrov at fastmail dot fm
 You can use it almost freely but
 1: don't [re]move that comment block from sources.
 2: if you distibute this code,or derived works in binary form,put somewhere
 information that i wrote it.
 No warranty at all,use that at your own risk.
*/
#ifndef _MATH3_H_
#define _MATH3_H_
static const double pi=(3.1415926535897932384626433832795);
template <class T> struct vec3{
        public:
                T x,y,z;
                vec3(T a,T b,T c){
                        x=a;y=b;z=c;
                }
                vec3(){
                }
                inline vec3<T> operator +(const vec3<T> &b)const{
                        vec3<T> result;
                        result.x=x+b.x;
                        result.y=y+b.y;
                        result.z=z+b.z;
                        return result;
                }
                inline vec3<T> operator -(const vec3<T> &b)const{
                        vec3<T> result;
                        result.x=x-b.x;
                        result.y=y-b.y;
                        result.z=z-b.z;
                        return result;
                }
                                inline vec3<T> operator -()const{
                        vec3<T> result;
                        result.x=-x;
                        result.y=-y;
                        result.z=-z;
                        return result;
                }
                inline vec3<T> operator *(const T &b) const {
                        vec3<T> result;
                        result.x=x*b;
                        result.y=y*b;
                        result.z=z*b;
                        return result;
                }
                inline vec3<T> operator /(const T &b)const{
                        vec3<T> result;
                        T a=1/b;
                        result.x=x*a;
                        result.y=y*a;
                        result.z=z*a;
                        return result;
                }
                inline vec3<T>& operator +=(const vec3<T> &b){
                        x+=b.x;
                        y+=b.y;
                        z+=b.z;
                        return *this;
                }
                inline vec3<T>& operator -=(const vec3<T> &b){

                        x-=b.x;
                        y-=b.y;
                        z-=b.z;
                        return *this;
                }
                inline vec3<T>& operator *=(const T &b){
                        x*=b;
                        y*=b;
                        z*=b;
                        return *this;
                }
                inline vec3<T>& operator /=(const T &b){
                        T a=1/b;
                        x*=a;
                        y*=a;
                        z*=a;
                        return *this;
                }
};

/*template<class A> inline vec3<T>& operator=(const vec3<A> &b){
                                        x=b.x;
                                        y=b.y;
                                        z=b.z;
                                        return *this;
                                }
                                */

template <class T> vec3<T> CrossProd(const vec3<T> &a,const vec3<T> &b){
        vec3<T> result;
        result.x=a.y*b.z - a.z*b.y;
        result.y=a.z*b.x - a.x*b.z;
        result.z=a.x*b.y - a.y*b.z;
        return result;
};

template <class T> inline T DotProd(const vec3<T> &a,const vec3<T> &b){
        return a.x*b.x + a.y*b.y + a.z*b.z;
};

template <class T> inline T Length2(const vec3<T> &a){
        return a.x*a.x + a.y*a.y + a.z*a.z ;
};

template <class T> inline T Length(const vec3<T> &a){
        return sqrt( a.x*a.x + a.y*a.y + a.z*a.z );
};

template <class T> inline void Normalize (vec3<T> &a){
        T il=1/sqrt( a.x*a.x + a.y*a.y + a.z*a.z );
        a*=il;
};

template <class T> inline vec3<T> Normalized (const vec3<T> &a){
        T il=1/sqrt( a.x*a.x + a.y*a.y + a.z*a.z );
        return (a*il);
};

template <class T> struct quaternion{
        T a,b,c,d;

        inline quaternion(const vec3<T> &axis, T angle){
                vec3<T> axisn=Normalized(axis);
                a=cos(angle/2);
                T s=sin(angle/2);
                b=axisn.x*s;
                c=axisn.y*s;
                d=axisn.z*s;
        }
        inline quaternion(T _a,T _b,T _c,T _d){
                a=_a;b=_b;c=_c;d=_d;
        }
        inline quaternion(){
        }
        inline quaternion<T> operator *(const quaternion<T> &q2) const{
                quaternion<T> result;
                //
                result.a = a*q2.a - b*q2.b - c*q2.c - d*q2.d;
                result.b = a*q2.b + b*q2.a + c*q2.d - d*q2.c;
                result.c = a*q2.c - b*q2.d + c*q2.a + d*q2.b;
                result.d = a*q2.d + b*q2.c - c*q2.b + d*q2.a;
                return result;
        };
        inline quaternion<T> operator *(T s) const{
                quaternion<T> result;
                // borrowed from some textbook i can't remember now.
                result.a = a*s;
                result.b = b*s;
                result.c = c*s;
                result.d = d*s;
                return result;
        };
        inline quaternion<T>& operator *=(T s){
                // borrowed from some textbook i can't remember now.
                a *= s;
                b *= s;
                c *= s;
                d *= s;
                return *this;
        };
};

template <class T> inline T Length2(const quaternion<T> &q){
        return  q.a*q.a + q.b*q.b + q.c*q.c + q.d*q.d ;
};

template <class T> inline T Length(const quaternion<T> &q){
        return sqrt( q.a*q.a + q.b*q.b + q.c*q.c + q.d*q.d );
};
template <class T> inline void Normalize(quaternion<T> &q){
        T il=1/Length(q);
        q*=il;
};

template <class T> inline quaternion<T> Normalized(const quaternion<T> &q){
        T il=1/Length(q);
        return q*il;
};

//--- does inverse transform
template <class T> inline quaternion<T> Inverse(const quaternion<T> &q){
        return quaternion<T>(-q.a , q.b , q.c , q.d);
};

template <class T> struct matrix3x3{
        T a[3][3];
        inline matrix3x3(){
        };

        inline matrix3x3(T m00,T m01,T m02,T m10,T m11,T m12,T m20,T m21,T m22){
                a[0][0]=m00;
                a[0][1]=m01;
                a[0][2]=m02;
                a[1][0]=m10;
                a[1][1]=m11;
                a[1][2]=m12;
                a[2][0]=m20;
                a[2][1]=m21;
                a[2][2]=m22;
        };

        inline matrix3x3<T> operator*(const matrix3x3<T> &m2)const{
                matrix3x3<T> result;

                result.a[0][0] = a[0][0]*m2.a[0][0] + a[0][1]*m2.a[1][0] + a[0][2]*m2.a[2][0] ;
                result.a[0][1] = a[0][0]*m2.a[0][1] + a[0][1]*m2.a[1][1] + a[0][2]*m2.a[2][1] ;
                result.a[0][2] = a[0][0]*m2.a[0][2] + a[0][1]*m2.a[1][2] + a[0][2]*m2.a[2][2] ;
                result.a[1][0] = a[1][0]*m2.a[0][0] + a[1][1]*m2.a[1][0] + a[1][2]*m2.a[2][0] ;
                result.a[1][1] = a[1][0]*m2.a[0][1] + a[1][1]*m2.a[1][1] + a[1][2]*m2.a[2][1] ;
                result.a[1][2] = a[1][0]*m2.a[0][2] + a[1][1]*m2.a[1][2] + a[1][2]*m2.a[2][2] ;
                result.a[2][0] = a[2][0]*m2.a[0][0] + a[2][1]*m2.a[1][0] + a[2][2]*m2.a[2][0] ;
                result.a[2][1] = a[2][0]*m2.a[0][1] + a[2][1]*m2.a[1][1] + a[2][2]*m2.a[2][1] ;
                result.a[2][2] = a[2][0]*m2.a[0][2] + a[2][1]*m2.a[1][2] + a[2][2]*m2.a[2][2] ;

                return result ;
        };
        inline vec3<T> operator*(const vec3<T> &v){
                vec3<T> result;
                result.x = a[0][0]*v.x+a[0][1]*v.y+a[0][2]*v.z;
                result.y = a[1][0]*v.x+a[1][1]*v.y+a[1][2]*v.z;
                result.z = a[2][0]*v.x+a[2][1]*v.y+a[2][2]*v.z;
                return result;
        }
        inline T& c(int i,int j) {
                return (a[((i+3000000) % 3)][((j+3000000) % 3)]);
        };
        inline T& cf(int i,int j) {
                return (a[i][j]);
        };
};

template <class T> T Det(matrix3x3<T> &m){
        int i,j;
        T b,c,r=0;
        for(j=0;j<3;++j){
                b=m.c(0,j);
                c=b;
                for(i=1;i<3;++i){
                        b*=m.cf(i,(j+i)%3);
                        c*=m.cf(i,(j-i+3)%3);
                };
                r+=b-c;
        };
        return r;
}
template <class T> inline T Adj(matrix3x3<T> &m,int i,int j){
        return m.c(i+1,j+1)*m.c(i+2,j+2)-m.c(i+1,j-1)*m.c(i+2,j-2);
}

template <class T> matrix3x3<T> Inverse(matrix3x3<T> &m){
        int i,j;
        T d=Det(m);
        if(d!=0){d=1/d;} else{
                std::cout<<"error"<<endl;
        };
        matrix3x3<T> result;
        for(i=0;i<3;++i){
                for(j=0;j<3;j++){
                        result.cf(i,j)=Adj(m,j,i)*d;
                }
        };
        return result;
};
template <class T> matrix3x3<T> QuaternionToMatrix(const quaternion<T> &q){

        T sq[4]={q.a*q.a , q.b*q.b , q.c*q.c , q.d*q.d};
    /*    matrix3x3<T> result;
        result.c(0,0)=  sq[0]+sq[1]-sq[2]-sq[3];        result.c(0,1)=  2*(q.b*q.c-q.a*q.d);            result.c(0,2)=2*(q.a*q.c+q.b*q.d);
    result.c(1,0)=2*(q.a*q.d+q.b*q.c);          result.c(1,1)=sq[0]+sq[2]-sq[1]-sq[3];  result.c(1,2)=2*(q.c*q.d-q.a*q.b);
    result.c(2,0)=2*(q.b*q.d-q.a*q.c);                  result.c(2,1)=2*(q.a*q.b+q.c*q.d);              result.c(2,2)=sq[0]+sq[3]-sq[1]-sq[2];
        */
                matrix3x3<T> result(

                        sq[0]+sq[1]-sq[2]-sq[3],        2*(q.b*q.c-q.a*q.d),            2*(q.a*q.c+q.b*q.d),
                        2*(q.a*q.d+q.b*q.c),            sq[0]+sq[2]-sq[1]-sq[3],        2*(q.c*q.d-q.a*q.b),
                        2*(q.b*q.d-q.a*q.c),            2*(q.a*q.b+q.c*q.d),            sq[0]+sq[3]-sq[1]-sq[2]

                        );

        return result;
};

template <class T> struct CoordSys{
        quaternion<T> orientation;
                vec3<T> position;
        CoordSys(){
        };
        CoordSys(quaternion<T> o,vec3<T>pos){
                position=pos;
                orientation=o;
        };
                //helpser
                void Translate(const vec3<T> &by){
                        position+=QuaternionToMatrix(orientation)*by;
                };
                void Rotate(const quaternion<T> &by){
                        orientation=
#ifndef dont_normalize
                        Normalized
#endif
                                (orientation*by);
                };
                inline void Rotate(const vec3<T> &axis,T angle){
                        orientation=
#ifndef dont_normalize
                        Normalized
#endif
                                (orientation*(quaternion<T>(axis,angle)));
                }
};
template<class T> CoordSys<T> operator*(const CoordSys<T> &c1,const CoordSys<T> &c2){
        CoordSys<T> result;
        result.orientation=
#ifndef dont_normalize
                        Normalized
#endif
                        (c1.orientation*c2.orientation)
                        ;
        result.position=QuaternionToMatrix(c1.orientation)*c2.position+c1.position;
};

template<class T> CoordSys<T> Inverse(const CoordSys<T> &c1){
        CoordSys<T> result;
        result.orientation=Inverse(c1.orientation);
        result.position= -((QuaternionToMatrix(result.orientation))*c1.position);
        return result;
};

template<class T> std::ostream& operator<<(std::ostream&s,const vec3<T> &v){
        s<<"vector:"<<std::endl;
        s<<"["<<v.x<<",\t"<<v.y<<",\t"<<v.z<<"]"<<std::endl;
        return s;
};
template<class T> std::ostream& operator<<(std::ostream&s,const quaternion<T> &q){
        s<<"quaternion:"<<std::endl;
        s<<"["<<q.a<<",\t"<<q.b<<",\t"<<q.c<<",\t"<<q.d<<"]"<<std::endl;
        return s;
};
template<class T> std::ostream& operator<<(std::ostream&s,const matrix3x3<T> &m){
        s<<"matrix:"<<std::endl;
        for(int i=0;i<3;i++){
                s<<"[";
                for(int j=0;j<3;j++){
                        s<<m.a[i][j];
                        if(j<2)s<<",\t";
                };
                s<<"]"<<std::endl;
        };
        return s;
};


typedef vec3<float> vec3f;
typedef vec3<double> vec3d;
typedef quaternion<float> quaternionf;
typedef quaternion<double> quaterniond;
typedef matrix3x3<float> matrix3x3f;
typedef matrix3x3<double> matrix3x3d;
typedef CoordSys<float> CoordSysf;
typedef CoordSys<double> CoordSysd;

#endif //header